#include <opencv2/opencv.hpp>
#include <opencv2/ml.hpp>
#include <fstream>
#include <vector>
#include <cmath>
#include <iostream>

struct BoundingBox {
    std::string filename;  // 文件名
    float label;           // 标签
    // save x1, y1, x2, y2 to rect
    cv::Rect rect;
    float x1, y1, x2, y2;  // 左上角坐标和右下角坐标
    float confidence;      // 置信度
};

cv::Mat create_density_map(const cv::Mat& im0, const std::vector<BoundingBox>& boxes, const std::string& pname, 
                           const int k = 5, const float sigma = 35, const int kernel_size = 201, const float alpha = 0.5, const float beta = 1500) {
    int h = im0.rows;
    int w = im0.cols;
    printf("h = %d, w = %d\n", h, w);
    cv::Mat density_map = cv::Mat::zeros(h, w, CV_32F);

    // 计算中点
    std::vector<cv::Point2f> middle_points;
    std::vector<float> cls, conf;
    std::vector<cv::Rect> xyxy;
    for (const auto& box : boxes) {
        cv::Point2f middle_point((box.x1 + box.x2) / 2, (box.y1 + box.y2) / 2);
        middle_points.push_back(middle_point);
        cls.push_back(box.label);
        conf.push_back(box.confidence);
        xyxy.push_back(box.rect);
    }

    // 创建并训练 KD 树
    cv::Ptr<cv::ml::KNearest> kd_tree = cv::ml::KNearest::create();
    cv::Mat samples(middle_points.size(), 2, CV_32F, middle_points.data());

    std::vector<int> cls_int(cls.begin(), cls.end());  // 将 float 转换为 int
    cv::Mat labels(cls_int.size(), 1, CV_32S, cls_int.data());
    kd_tree->train(samples, cv::ml::ROW_SAMPLE, labels);

    // 查询 KD 树
    cv::Mat results, neighbours, dists;
    kd_tree->findNearest(samples, k, results, neighbours, dists);

    for (size_t i = 0; i < xyxy.size(); ++i) {
        const cv::Rect& rect = xyxy[i];
        int cls_i = cls[i];
        float conf_i = conf[i];
        float area = rect.width * rect.height;

        // 计算得分
        float score = 0;
        for (int j = 1; j < k; ++j) {
            score += dists.at<float>(i, j) * 1e-4;
        }
        if (score > 100) score = 100;

        // 创建 mask
        cv::Mat mask = cv::Mat::zeros(h, w, CV_32F);
        float value = (area) / (std::sqrt(score)+alpha);
        mask.at<float>((rect.y + rect.height/2), (rect.x + rect.width/2)) = value;
        // print rect.y + rect.height, rect.x + rect.width
        printf("(%d, %d) area = %f, score = %f\n, value = %f\n", rect.y + rect.height/2, rect.x + rect.width/2, area, score, value);

        // 累加到密度图
        density_map += mask;
    }

    // 高斯模糊
    cv::GaussianBlur(density_map, density_map, cv::Size(kernel_size, 201), sigma);
    density_map *= beta;

    // 截断
    density_map.setTo(255, density_map > 255);
    density_map.setTo(0, density_map < 0);

    // 转换为 8 位图像
    cv::Mat density_map_8u;
    density_map.convertTo(density_map_8u, CV_8U);

    // 应用颜色映射
    cv::Mat density_map_color;
    cv::applyColorMap(density_map_8u, density_map_color, cv::COLORMAP_JET);

    // 叠加到原图
    cv::Mat output;
    cv::addWeighted(im0, 0.5, density_map_color, 0.5, 0, output);

    // 保存结果
    cv::imwrite("./" + pname, output);

    return density_map_color;
}

// read txt file return a vector of BoundingBox
std::vector<BoundingBox> read_txt(const std::string& filename) {
    std::vector<BoundingBox> boxes;
    std::ifstream fin(filename);
    // format of the file: filename label x1 y1 x2 y2 confidence
    // 12.jpg 0 342 166 359 185 0.252197
    std::string line;
    while (std::getline(fin, line)) {
        std::istringstream iss(line);
        std::string filename;
        float label, x1, y1, x2, y2, confidence;

        iss >> filename >> label >> x1 >> y1 >> x2 >> y2 >> confidence;

        BoundingBox box;
        box.filename = filename;
        box.label = label;
        box.x1 = x1;
        box.y1 = y1;
        box.x2 = x2;
        box.y2 = y2;
        box.rect = cv::Rect(x1, y1, x2 - x1, y2 - y1);
        box.confidence = confidence;

        boxes.push_back(box);
    }
    return boxes;
}

int main() {
    // 示例数据
    // define the path of image
    std::string path = "/home/chenl/project/cv/crowd/YOLO-CROWD/data/head-count/heads/test/images/";
    std::string name = "12.jpg";
    cv::Mat im0 = cv::imread(path + name);
    std::string pname = "output.jpg";

    // read xyxy from xyxy_conf.txt
    std::string txt_path = "/home/chenl/project/cv/crowd/density_map_cpp/src/xyxy_conf12.txt";
    std::vector<BoundingBox> boxes = read_txt(txt_path);

    // 调用函数
    // parameters: 
    // im0: input image
    // boxes: vector of BoundingBox
    // pname: output image name
    // k: number of neighbours for KD tree, 调节人间距对密度图的影响,k越大,密度值越低
    int k = 5;
    // simga: sigma for Gaussian blur, 高斯模糊参数, 调节模糊程度,sigma越大,模糊程度越高
    float sigma = 35;
    // kernel_size: kernel size for Gaussian blur, 高斯模糊核大小, 调节模糊程度,kernel_size越大,覆盖范围越广,模糊程度越高
    int kernel_size = 201;
    // alpha: parameter for score calculation, 得分计算参数, 调节人群密度对得分的影响,alpha越大,密度值越低
    float alpha = 0.5;
    // beta: parameter for density map scaling, 密度图缩放参数, 调节密度图的整体强度,beta越大,密度值越搞
    float beta = 1500;
    cv::Mat density_map = create_density_map(im0, boxes, pname, k, sigma, kernel_size, alpha, beta);

    // 显示结果
    //cv::imshow("Density Map", density_map);
    cv::waitKey(0);

    return 0;
}